/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
  * © 2016 OX Software GmbH, Germany. info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/tk/control/spinfield', [
    'io.ox/office/tk/utils',
    'io.ox/office/tk/keycodes',
    'io.ox/office/tk/forms',
    'io.ox/office/tk/utils/tracking',
    'io.ox/office/tk/control/textfield',
    'io.ox/office/settings/units'
], function (Utils, KeyCodes, Forms, Tracking, TextField, Units) {

    'use strict';

    // class SpinField ========================================================

    /**
     * Creates a numeric field control with additional controls used to spin
     * (decrease and increase) the value.
     *
     * @constructor
     *
     * @extends TextField
     *
     * @param {Object} [initOptions]
     *  Optional parameters. Supports all options of the TextField base class.
     *  If the option 'validator' exists, it MUST be an instance of
     *  TextField.NumberValidator (or a sub class). Alternatively, the options
     *  supported by the constructor of the class TextField.NumberValidator can
     *  be passed. In this case, a new instance of TextField.NumberValidator
     *  will be created automatically based on these options. Additionally, the
     *  following options are supported:
     */
    function SpinField(initOptions) {

        var // self reference
            self = this,

            // the number validator of this spin field
            validator = Utils.getObjectOption(initOptions, 'validator') || new TextField.NumberValidator(initOptions),

            // whether to turn of spin controls
            spinEnabled = Utils.getBooleanOption(initOptions, 'spinEnabled', true);

        // base constructors --------------------------------------------------

        TextField.call(this, Utils.extendOptions({ keyboard: 'number' }, initOptions, { validator: validator }));

        // private methods ----------------------------------------------------

        function getSmallStep() {
            return Units.getStandardUnitStep();
        }

        function getLargeStep() {
            return getSmallStep() * 5;
        }

        /**
         * Changes the current value of the spin field, according to the passed
         * step of the general unit.
         */
        function changeValue(step) {

            var //
                unit = Units.getStandardUnit(),
                // the current value of the spin field
                oldValue = Utils.convertHmmToLength(self.getFieldValue(), unit),
                // the new value
                newValue = oldValue + step;

            newValue = Utils.convertLengthToHmm(newValue, unit);
            // set new value, after restricting to minimum/maximum
            newValue = validator.restrictValue(newValue);
            self.setValue(newValue);
        }

        /**
         * Creates and returns a spin button. Initializes tracking used to
         * change the current value.
         */
        function createSpinButton(increase) {

            var // create the spin button element
                spinButton = $(Forms.createButtonMarkup({ focusable: false })).addClass(increase ? 'spin-up' : 'spin-down');

            // enables or disables node tracking according to own enabled state
            function initializeTracking() {
                if (self.isEnabled()) {
                    Tracking.enableTracking(spinButton, { autoRepeat: true });
                } else {
                    Tracking.disableTracking(spinButton);
                }
            }

            // add the caret symbol
            spinButton.append(Forms.createCaretMarkup(increase ? 'up' : 'down'));

            // enable/disable node tracking according to own enabled state
            self.on('group:enable', initializeTracking);
            initializeTracking();

            // register tracking event listeners to change the current value
            spinButton.on({
                'tracking:start tracking:repeat': function () { changeValue(increase ? getSmallStep() : -getSmallStep()); },
                'tracking:end tracking:cancel': function () { self.executeDelayed(function () { self.getInputNode().focus(); }, undefined, 'TK: createSpinButton'); }
            });

            // enable/disable the button according to the current value
            self.registerUpdateHandler(function (value) {
                var enabled = increase ? (value < validator.getMax()) : (value > validator.getMin());
                Forms.enableNodes(spinButton, enabled);
            });

            return spinButton;
        }

        /**
         * Handles keyboard events in the text field.
         */
        function inputKeyDownHandler(event) {

            if (KeyCodes.hasModifierKeys(event)) { return; }

            switch (event.keyCode) {
                case KeyCodes.UP_ARROW:
                    changeValue(getSmallStep());
                    return false;
                case KeyCodes.DOWN_ARROW:
                    changeValue(-getSmallStep());
                    return false;
                case KeyCodes.PAGE_UP:
                    changeValue(getLargeStep());
                    return false;
                case KeyCodes.PAGE_DOWN:
                    changeValue(-getLargeStep());
                    return false;
            }
        }

        /**
         * Handles updates of the input text field
         */
        function updateHandler(value) {
            this.getInputNode().attr({ 'aria-valuenow': value, 'aria-valuetext': this.getInputNode().val() });
        }

        // initialization -----------------------------------------------------

        if (spinEnabled) {
            // add special marker class used to adjust formatting, add spin buttons
            this.getNode().addClass('spin-field').append(
                $('<div>').addClass('spin-wrapper').append(createSpinButton(true), createSpinButton(false))
            );

            // add ARIA attributes (note: role 'spinbutton' would be correct but doesn't work with VoiceOver)
            this.getInputNode().attr({ role: 'slider', 'aria-valuemin': validator.getMin(), 'aria-valuemax': validator.getMax() });

            // register event handlers
            this.getInputNode().on('keydown', inputKeyDownHandler);
        }

        // register input field update handler
        this.registerUpdateHandler(updateHandler);

    } // class SpinField

    // exports ================================================================

    // derive this class from class TextField
    return TextField.extend({ constructor: SpinField });

});
